/* cc_2915.c: KineticSystems Corporation 2915 PCI Card Driver */
/*
 *  filename: cc_2915.c
 *  version : 2.00
 *  date    : 11-Jun-1997
 *  modified: 11-Aug-1997
 *            11-Oct-1997  DMA bug fixed by Y.Tanaka
 *             5-Nov-1997  fixed Set Crate number problem by Y.Tanaka 
 *                         DMA mode QREPEAT
 *             7-Nov-1997  fixed DMA transfer size in byte by Y.Tanaka
 *             3-Mar-1998  features added, problems fixed by C.Timmer :
 *			   1) fixed wrong x & q return values
 *			   2) return x,q, AND error data for camac_single
 *			   3) fixed addressing a non-existing crate causing
 *			      error on next command to existing crate
 *			   4) DMA transfer extended to all modes (untested)
 *			   5) parallel poll capability added
 *			   6) access to status of last command added
 *			   7) dump of registers added
 *			   8) can share IRQ if necessary
 *			   9) separate the control of lam enable/disable
 *			      from pci lam interrupt enable/disable
 *			  10) set pci latency timer to recommended value
 *			  11) only look for one KS2915 card.
 *
 *  author  : M.Haseno, Y.Tanaka
 *  modified: Y.Nagasaka
 *            Nagasaki Institute of Applied Science (NIAS)
 *  modified: C.Timmer
 *            Thomas Jefferson National Accelerator Facility
 *
 */
#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/major.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/pci.h>
#include <linux/bios32.h>

#include <asm/io.h>
#include "cc_2915.h"

#define BUFFER_SIZE 512*1024     /* DMA Buffer Size 512KBytes */
#ifdef DEBUG
  #define DEBUG2915 1
#else
  /* #define DEBUG2915 1 */
#endif

static volatile char cc_buf[BUFFER_SIZE]; 
static char device[]="ks2915"; 
static int try, done_try;

struct cc_struct {
  int base1;
  int base2;
  unsigned int irq;
  int status;
  short cur_branch;
  short cur_crate;
  struct wait_queue *cc_wait_q;
};

struct cc_struct cc_table[] =
{
  {0, 0, 0, CC_K_READY, 0, 0, NULL,}
};

struct  camac {
  /* Parallel Bus Operational Register */
  unsigned int csr;
  unsigned int cnaf;
  unsigned int tcr;
  unsigned int srr;
  /* PCI Interface Operational Register */
  unsigned int fifo;
  unsigned int mwa;
  unsigned int mwtc;
  unsigned int mra;
  unsigned int mrtc;
  unsigned int ics;
  unsigned int bmcs;  
};

struct camac reg[] =
{
  {CSR, CNAF, TCR, SRR, FIFO, MWA, MWTC, MRA, MRTC, ICS, BMCS,},
};

struct naf_data {
  short mode;
  short naf;
  unsigned int data;
  unsigned int xq;
};

static int cc_probe(struct cc_struct *cc_pci_hdr, struct camac *cc_reg);

static void
cc_interrupt (int irq, void *dev_id, struct pt_regs *regs)
{
  struct cc_struct *cc = cc_table;
  unsigned int csr;
  
  if (dev_id != device) {
    printk("cc_interrupt: wrong device, dev_id = %p\n", dev_id);
    return;
  }
  
  if (irq != cc->irq) {
    printk("cc_interrupt: wrong irq (=%d)\n",irq);
    return;
  }

  csr = inl(reg->csr);
  
  /*
   * Reject improper interrupts. These happen when sharing
   * IRQ with another device - hardware problems?
   */
  if (!((csr & RFS) && (csr & RFS_IENA))) {
    return;
  }

  outl((unsigned int) 
      ((~RFS_IENA & csr & RFS_MASK) | CLR_DNI | CLR_PCII), reg->csr);

#ifdef DEBUG2915
  printk("cc_interrupt: csr=%#x\n", csr);
#endif
 
  wake_up_interruptible (&cc->cc_wait_q);
}

static void
camac_single (short mode, short naf, int *data)
{

#ifdef DEBUG2915
  printk("camac_single: naf = %#x, cnaf = %#x\n", naf, ((inl(reg->cnaf) & CRATE_MASK) | naf));
#endif

  outl((inl(reg->cnaf) & CRATE_MASK) | naf, reg->cnaf);

  if (mode == SINGLE16)
    outl( (unsigned int)SINGLE16 | ( inl(reg->csr) & WORD_MASK ), reg->csr);
  else
    outl( (unsigned int)SINGLE24 | ( inl(reg->csr) & WORD_MASK ), reg->csr);

  if ((CC_NAFRK_F16 & naf) && !(CC_NAFRK_F08 & naf)) {  /* Write Operation */
    try = 0;
    while(1){
      if(inl(reg->bmcs) & OUT_FIFO_FULL) {
	try++ ;
	if(try > MAX_FIFO_TRY) {
	  printk("camac_single: Timeout, outbound FIFO full\n");
	  break;
	}
      }
      else {
	outl((unsigned int) (*data), reg->fifo);
        outl((unsigned int) (GO | (inl(reg->csr) & CSR_MASK)), reg->csr);
	done_try = 0;
	while (!(inl(reg->csr) & DONE)){
	  done_try++;
	  if(done_try > MAX_DONE_TRY) {
	    printk("camac_single: Timeout, Write waiting for DONE\n");
	    break;
	  }
	}
	break;
      }
    } 
  }

  else if (!(CC_NAFRK_F16 & naf) && !(CC_NAFRK_F08 & naf)) { /* Read Operation */

    outl((unsigned int) (GO |(inl(reg->csr) & CSR_MASK)), reg->csr);
    done_try = 0;
    while (!(inl(reg->csr) & DONE)){
      done_try++;
      if(done_try > MAX_DONE_TRY) {
	printk("camac_single: Timeout, Read waiting for DONE\n");
	break;
      }
    }

    try = 0;
    while(1){
      if(inl(reg->bmcs) & IN_FIFO_EMPTY) {
	try++ ;
	if(try > MAX_FIFO_TRY) {
	  printk("camac_single: Timeout, inbound FIFO empty\n");
	  break;
	}
      }
      else {
	*data = inl(reg->fifo);
	break;
      }
    }  
  }

  else { /* Operational Commands */

    outl((unsigned int) (GO | inl(reg->csr)), reg->csr);
    done_try = 0;
    while (!(inl(reg->csr) & DONE)){
      done_try++;
      if(done_try > MAX_DONE_TRY) {
	printk("camac_single: Timeout, Operation waiting for DONE\n");
	break;
      }
    }
  }
}

static void
cc_reset (void)
{
  int i = CLR;
  
  outl((unsigned int)  RST_INFC, reg->csr);
  outl((unsigned int) (CLR_PCII   | inl(reg->csr)),  reg->csr);
  outl((unsigned int) (FIFO_RESET | inl(reg->bmcs)), reg->bmcs);
   
  done_try = 0;
  while (!(inl(reg->csr) & DONE)){
    done_try++;
    if(done_try > MAX_DONE_TRY) {
	printk("cc_reset: Timeout, waiting for DONE\n");
	break;
    }
  }
  camac_single(SINGLE16, NAFGEN(30, 0, 17), &i);  /* Crate Controller Reset */

#ifdef DEBUG2915
  printk("cc_reset: Crate Reset Complete, csr=%#x bmcs=%#x\n", inl(reg->csr),inl(reg->bmcs));
#endif

}

static int
cc_read (struct inode *inode, struct file *file, char *buffer, int size)
{

  int sizeb;
  
#ifdef DEBUG2915
  int i;
  printk ("cc_read: enter with count=%d\n", size);
#endif

  if(inl(reg->csr) & SINGLE16) {
    sizeb = 2*size;
    outl(~((sizeb>>1)-1), reg->tcr);
  }
  else {
    sizeb = 4*size;
    outl(~((sizeb>>2)-1), reg->tcr);
  }
  outl((unsigned long) cc_buf, reg->mwa);
  outl(sizeb, reg->mwtc);

#ifdef DEBUG2915
  printk("cc_read: cc_buf=%x,  mwa=%x \n",(unsigned int)cc_buf, inl(reg->mwa));
  printk("cc_read: tcr=%x,  mwtc=%x\n",(unsigned int)inl(reg->tcr),inl(reg->mwtc));
#endif

  outl((unsigned int) (FIFO_RESET | inl(reg->bmcs)), reg->bmcs); 
  outl((unsigned int) (GO | inl(reg->csr)), reg->csr);

  /* Enable DMA from 2915 to PCI mem */
  outl(WTT_ENA, reg->bmcs);

  done_try = 0;
  while (!(inl(reg->csr) & DONE)){
    done_try++;
    if(done_try > MAX_DMA_TRY) {
      printk("cc_read: Timeout, DMA Read waiting for DONE\n");
      break;
    }
  }

  /* Set Single mode */
  outl(0xFFF0 & inl(reg->csr), reg->csr);

#ifdef DEBUG2915
  printk("cc_read: csr=%x,  bmcs=%x\n", inl(reg->csr), inl(reg->bmcs));
  printk("  after DMA: cc_buf=%x,  mwa=%x \n",(unsigned int)cc_buf, inl(reg->mwa));
  printk("  after DMA: tcr=%x,  mwtc=%x\n",inl(reg->tcr),inl(reg->mwtc));
#endif

  outl( (unsigned int)(FIFO_RESET | (~WTT_ENA & inl(reg->bmcs))), reg->bmcs);
  if(inl(reg->csr) & ERR)
    return -E_CC_DMA;

#ifdef DEBUG2915
  for(i=0;i<10;i++)  printk("cc_buf[%d] = %x\n", i, (char)cc_buf[i]);
#endif

  memcpy_tofs((void *)buffer, (void *)cc_buf, sizeb);
  return 0;
}

static int
cc_write (struct inode *inode, struct file *file, const char *buffer, int size)
{

  int sizeb;
#ifdef DEBUG2915  
  printk ("cc_write: enter with count=%d\n", size);
#endif

  /*
   * reg->tcr specifies max # of words to be transferred from
   * 2915 to the crate.
   * reg->mrtc specifies # bytes to be transferred from PCI bus
   * to 2915.
   * 
   * reg->tcr format in 2's complement. Note that the >>1 and
   * >>2 simply divide by 2 and 4 to get the # of words.
   */
  if(inl(reg->csr) & SINGLE16) {
    sizeb = 2*size;
    outl(~((sizeb>>1)-1), reg->tcr);
  }
  else {
    sizeb = 4*size;
    outl(~((sizeb>>2)-1), reg->tcr);
  }
  
  /* Moved this line from end of routine to here. (timmer) */
  memcpy_fromfs((void *)cc_buf, (void *)buffer, sizeb);

  /*
   * static volatile char cc_buf[BUFFER_SIZE]
   * The address of this array is put in reg->mra.
   * The num of bytes is put in reg->mrtc.
   */
  
  outl((unsigned long) cc_buf, reg->mra);
  outl(sizeb, reg->mrtc);

  /* Reset in and outbound fifos & go */
  outl((unsigned int) (FIFO_RESET | inl(reg->bmcs)), reg->bmcs); 
  outl((unsigned int) (GO | inl(reg->csr)), reg->csr);

  /* Enable DMA from PCI mem to 2915 */
  outl(RDT_ENA, reg->bmcs);

  done_try = 0;
  while (!(inl(reg->csr) & DONE)){
    done_try++;
    if(done_try > MAX_DMA_TRY) {
      printk("cc_write: Timeout, DMA Write waiting for DONE\n");
      break;
    }
  }

  /* Set Single mode */
  outl(0xFFF0 & inl(reg->csr), reg->csr);
  
  outl( (unsigned int)(FIFO_RESET | (~RDT_ENA & inl(reg->bmcs))), reg->bmcs);
  if(inl(reg->csr) & ERR)
    return -E_CC_DMA;
    
  return 0; 
}

static int
cc_ioctl (struct inode *inode, struct file *file,
	  unsigned int cmd, unsigned long arg)
{
  unsigned int minor = MINOR (inode->i_rdev);
  struct cc_struct *cc = &cc_table[minor];
  struct naf_data command;
  int i, retval = 0;
  
#ifdef DEBUG2915
  unsigned int ui;
  printk("cc_ioctl: Entered cmd = %x \n",cmd);
#endif

  switch (cmd) {
    
    case CCIOC_RESET:           /* Reset Interface */

      cc_reset();
      #ifdef DEBUG2915
        printk("CCIOC_RESET: reset\n");
      #endif
      break;
    
    case CCIOC_SET_BRANCH:      /* Set Branch */

      if (arg+1 > 0 && arg < CC_K_MAX_BRANCH)
        cc->cur_branch = arg;
      else
        return -EINVAL;
      #ifdef DEBUG2915
        printk("CCIOC_SET_BRANCH: crate branch to %hd\n", cc->cur_branch);
      #endif
      break;
    
    case CCIOC_CHK_BRANCH:      /* Check Branch */

      put_user(cc->cur_branch, (int *) arg);
      #ifdef DEBUG2915
        printk("CCIOC_CHK_BRANCH: crate branch = %hd\n", cc->cur_branch);
      #endif
      break;
    
    case CCIOC_SET_CRATE:       /* Set Crate */

      if (arg+1 > 0 && arg < CC_K_MAX_CRATE) {
        /* only do something if changing crates */
        if (cc->cur_crate != arg) {
          cc->cur_crate = (short) arg;
          outl((unsigned int) cc->cur_crate << 16, reg->cnaf);
          outl((unsigned int) (FIFO_RESET | inl(reg->bmcs)), reg->bmcs); 
        }
        else {
          #ifdef DEBUG2915
            printk("CCIOC_SET_CRATE: crate already set to %hd\n", cc->cur_crate);
          #endif
          break;
        }
      }
      else {
        return -EINVAL;
      }

      done_try = 0;
      while (!(inl(reg->csr) & DONE)){
        done_try++;
        if(done_try > MAX_DONE_TRY) {
	  printk("CCIOC_SET_CRATE: Timeout, waiting for DONE\n");
	  break;
        }
      }
      
      #ifdef DEBUG2915
        printk("CCIOC_SET_CRATE: crate set to %hd\n", cc->cur_crate);
      #endif

      break;
    
    case CCIOC_ENABLE_INT:

      #ifdef DEBUG2915
        printk("CCIOC_ENABLE_INT: enabled interrupt\n");
      #endif
      
      outl((unsigned int) 
	 (RFS_IENA | (inl(reg->csr) & RFS_MASK)), reg->csr);
      break;
    
    case CCIOC_DISABLE_INT:

      #ifdef DEBUG2915
        printk("CCIOC_DISABLE_INT: disabled interrupt\n");
      #endif
        
      outl((unsigned int) 
	 (~RFS_IENA & (inl(reg->csr) & RFS_MASK)), reg->csr);

      break;
    
    case CCIOC_ENABLE_LAM:

      camac_single(SINGLE16, NAFGEN (30, 0, 1), &i);
      i |= SRE;
      camac_single(SINGLE16, NAFGEN (30, 0, 17), &i);

      #ifdef DEBUG2915
        printk("CCIOC_ENABLE_LAM: Before cli cc_status: %x\n",i);
      #endif

      i = arg;
      camac_single(SINGLE24, NAFGEN (30, 13, 17), &i);

      #ifdef DEBUG2915
        printk("CCIOC_ENABLE_LAM: mask = %#x\n",i);
        printk("CCIOC_ENABLE_LAM: csr  = %#x\n",inl(reg->csr));
      #endif
      
      break;

    case CCIOC_DISABLE_LAM:

      camac_single(SINGLE16, NAFGEN (30, 0, 1), &i);
      i &= ~SRE;
      camac_single(SINGLE16, NAFGEN (30, 0, 17), &i);
    
      i = CLR;
      camac_single(SINGLE24, NAFGEN (30, 13, 17), &i);
      #ifdef DEBUG2915
        printk("CCIOC_DISABLE_LAM:\n");
      #endif
      break;
    
    case CCIOC_WAIT_LAM:

      #ifdef DEBUG2915
        printk("CCIOC_WAIT_LAM: timeout = %ld, csr = %04x\n", arg, inl(reg->csr));
      #endif

      /*
       * 'current' is a pointer to a task_struct structure which
       * is the current task on this processor.
       * Jiffies are in 100ths of a sec (see sched.h).
       */
      cli();
      if (arg == 0)
        current->timeout = jiffies + CC_K_TIMEOUT_LAM;
      else
        current->timeout = jiffies + arg;
        
      outl((unsigned int) 
	 (RFS_IENA | (inl(reg->csr) & RFS_MASK)), reg->csr);
      
      #ifdef DEBUG2915
        printk("Enter sleep %ld\n",current->signal);
      #endif
    
      interruptible_sleep_on (&cc->cc_wait_q);

      #ifdef DEBUG2915
        printk("Exit sleep %ld\n",current->signal);
      #endif

      if (current->signal & ~current->blocked)
        return -ERESTARTNOINTR;
    
      if (current->timeout == 0) {
        #ifdef DEBUG2915
          printk("CCIOC_WAIT_LAM: Timeout occurred\n");
        #endif
        return -ETIMEDOUT;
      }
      
      #ifdef DEBUG2915
        printk("CCIOC_WAIT_LAM: LAM occurred, timeleft = %ld\n",current->timeout);
      #endif
      break;
    
    case CCIOC_SINGLE:

      memcpy_fromfs(&command, (void *)arg, sizeof(command));
      camac_single(command.mode, command.naf, &command.data);
      command.xq = ((inl(reg->csr)) &
	(XQ_MASK | ERR | NAF_TIMEOUT | PBUS_TIMEOUT));
      memcpy_tofs((void *)arg, &command, sizeof(command));
    
      #ifdef DEBUG2915
        /* if mode = 16 bit, else 24 */
        if (command.mode)
          ui = (unsigned short) command.data;
        else
          ui = (unsigned int  ) command.data;
    
        printk("CCIOC_SINGLE: cmd mode=%#x, naf=%#x, data=%#x, cmd.xq=%#x, q,x=%d,%d\n",
           command.mode, command.naf, ui, command.xq, ((command.xq&NOQ)>>16)?0:1,((command.xq&NOX)>>17)?0:1);
        printk("\n");
      #endif
    
      break;
    
    case CCIOC_SETUP_DMA:

      memcpy_fromfs (&command, (void *)arg, sizeof(command));
      outl((inl(reg->cnaf) & CRATE_MASK) | command.naf, reg->cnaf);

      #ifdef DEBUG2915
        printk("CCIOC_SETUP_DMA: cmd mode=%#x, naf=%#x, data=%#x\n",
           command.mode, command.naf, command.data);
      #endif

      /* if 16bit words, else 24bit */
      if(command.mode & 0x01) 
        outl( SINGLE16 | inl(reg->csr), reg->csr);
      else
        outl(~SINGLE16 & inl(reg->csr), reg->csr);

      /* get transfer mode bits and set register */
      i = command.mode & TRANSFER_MASK;
      if ((i != QSTOP ) && (i != QIGNORE) &&
          (i != QSCAN ) && (i != QREPEAT) )
      {
        return -EINVAL;
      }
      outl(((inl(reg->csr) & ~TRANSFER_MASK) | i), reg->csr);
      break;
    
    case CCIOC_CMD_STAT:

      /* Check status of last cmd: x, q, error,& timeout bits. */
      i = ((inl(reg->csr)) &
	  (XQ_MASK | ERR | NAF_TIMEOUT | PBUS_TIMEOUT));
      put_user(i, (int *) arg);
 
      #ifdef DEBUG2915
        printk("CCIOC_CMD_STAT: error bits of csr reg = %x\n", i);
      #endif
    
      break;
    
    case CCIOC_PPOLL:

      /* Do a parallel poll. */    
      outl((unsigned int) (PP | GO |(inl(reg->csr) & CSR_MASK)), reg->csr);
      done_try = 0;
      while (!(inl(reg->csr) & DONE)){
        done_try++;
        if(done_try > MAX_DONE_TRY) {
	  printk("CCIOC_PPOLL: Timeout, waiting for poll\n");
	  break;
        }
      }

      i = inl(reg->srr);
      put_user(i, (int *) arg);

      #ifdef DEBUG2915
        printk("CCIOC_PPOLL: srr reg = %x\n", i);
      #endif
    
      break;
    
    case CCIOC_REG_DUMP:

      /* Print out all registers */
      {
        int r1,r2,r3,r4;
      
        r1=inl(reg->csr); r2=inl(reg->cnaf);
        r3=inl(reg->tcr); r4=inl(reg->srr);
        printk("CCIOC_DEBUG:\nParallel Bus Operational Registers\n");
        printk("   CSR       CNAF        TCR        SRR\n");
        printk("x%08x  x%08x  x%08x  x%08x\n",r1,r2,r3,r4);

        r1=inl(reg->bmcs); r2=inl(reg->ics);
        printk("PCI Interface Operational Registers\n");
        printk("   BMCS      ICS\n");
        printk("x%08x  x%08x\n",r1,r2);
      
        r1=inl(reg->mwa); r2=inl(reg->mwtc);
        r3=inl(reg->mra); r4=inl(reg->mrtc);
        printk("   MWA       MWTC        MRA        MRTC \n");
        printk("x%08x  x%08x  x%08x  x%08x\n",r1,r2,r3,r4);
      }
      break;
    
    default:
      retval = -EINVAL;
  }

  #ifdef DEBUG2915
    printk("cc_ioct: end\n");
  #endif

  return 0;
}

static int
cc_open (struct inode *inode, struct file *file)
{
  unsigned int minor = MINOR(inode->i_rdev);
  struct cc_struct *cc = &cc_table[minor];
  int retval=0;
  
  if (cc->status == CC_K_BUSY)
    return -EBUSY;

  /*
   * Here is the strategy. First try to use a fast, non-shared
   * interrupt. If that fails, it means another device is also
   * using this IRQ. Next, try to use a shared interrupt, if
   * that fails, you must either change the driver code of
   * the conflicting device so that it shares IRQs, or try to
   * get the bios to assign another IRQ to this card. Good luck!
   */
  if (request_irq(cc->irq, cc_interrupt,
		  SA_INTERRUPT, device ,(void *) device))
  {
    /* Try shared IRQ */
    if ((retval = request_irq(cc->irq, cc_interrupt,
 		  SA_SHIRQ, device ,(void *) device)))
    {
      printk("cc_open: can't allocate IRQ%d for K2915\n",cc->irq);
      return retval;
    } else {
      printk("cc_open: using shared interrupt\n");
    }
  } else {
    printk("cc_open: using fast interrupt\n");
  }
 
  cc_reset();
  cc->status = CC_K_BUSY;
  return 0;
}

static void
cc_release (struct inode *inode, struct file *file)
{
  unsigned int minor = MINOR(inode->i_rdev);
  struct cc_struct *cc = &cc_table[minor];

  free_irq(cc->irq, device);
  outl(0xFFF0 & inl(reg->csr), reg->csr); /* Set Single mode */
  /*  free_dma (cc->dma); */
  cc->status = CC_K_READY;
  return;
}

static struct file_operations cc_fops =
{
  NULL,                       /* lseek   */
  cc_read,
  cc_write,
  NULL,                       /* readdir */
  NULL,                       /* select  */
  cc_ioctl,
  NULL,                       /* mmap    */
  cc_open,
  cc_release,
};

int
cc_init (void)
{
  struct cc_struct *cc;
  int err;

  if (register_chrdev(K2915_MAJOR, "check_drv", &cc_fops)) {
    printk("cc_init: Cannot register major device %d for k2915\n", K2915_MAJOR);
    return -EIO;
  }

  if (!(err = cc_probe((struct cc_struct *)cc_table, (struct camac *)reg ))) {
    cc = cc_table;

    if ((check_region(cc->base1,IOR_SIZE)  != 0) ||
        (check_region(cc->base2,PBOR_SIZE) != 0))
    {
      printk("cc_init: No space for K2915 registers\n");
      return -EIO;
    }
    request_region(cc->base1,IOR_SIZE, "CAMAC K2915(IOR)" );
    request_region(cc->base2,PBOR_SIZE,"CAMAC K2915(PBOR)");
    
    printk("K2915 CAMAC Device Driver was Installed:\n");
    printk("   IRQ : 0x%8.8x\n",cc->irq);
    printk("   IOR : 0x%8.8x\n",cc->base1);
    printk("   PBOR: 0x%8.8x\n",cc->base2);
    printk("   CSR : 0x%8.8x\n",reg->csr);
    printk("   CNAF: 0x%8.8x\n",reg->cnaf);
    printk("   FIFO: 0x%8.8x\n",reg->fifo);

  }
  else
    printk("cc_init: BaseAddress or IRQ Error\n");
    
  return 0;  
}

#ifdef MODULE
int 
init_module (void)
{
  return cc_init();  
}

void
cleanup_module (void)
{
  struct cc_struct *cc = cc_table;
  if (MOD_IN_USE)
    printk("cc cleanup_module: K2915 busy (not removed)\n");
  else {
    unregister_chrdev(K2915_MAJOR, "check_drv");
    release_region(cc->base1,IOR_SIZE);
    release_region(cc->base2,PBOR_SIZE);
    printk("cc cleanup_module: K2915 device driver removed\n");
  }
}
#endif

static int 
cc_probe(struct cc_struct *cc_pci_hdr, struct camac *cc_reg)
{
  int pci_index, value = 0;
  unsigned char pci_bus, pci_device_fn, pci_irq_line, pci_timer;
  
  /* PCI Interface Operational Register (IOR) */
  unsigned int pci_ior_addr;    
  /* PCI Parallel Bus Operational Registers (PBOR) */
  unsigned int pci_pbor_addr;
  
  /*
   * This driver will only work with one card, so quit
   * after finding the first card and ignore the rest.
   */
  if (pcibios_present()) {
    for(pci_index = 0; pci_index < 16; pci_index++) {
      if (pcibios_find_device(PCI_VENDOR_ID_KINETIC,
			      PCI_DEVICE_ID_KINETIC_2915,
			      pci_index, &pci_bus, &pci_device_fn))
        break;

      /* One CANNOT set the IRQ here. It's set by the bios. */
      pcibios_read_config_byte(pci_bus, pci_device_fn, PCI_INTERRUPT_LINE, 
			       &pci_irq_line);	        
      pcibios_read_config_byte(pci_bus, pci_device_fn, PCI_LATENCY_TIMER, 
			       &pci_timer);
      if (pci_timer != LTR_SET) {
        pcibios_write_config_byte(pci_bus, pci_device_fn, PCI_LATENCY_TIMER,
                                  LTR_SET);
      }
                                
      pcibios_read_config_dword(pci_bus, pci_device_fn, PCI_BASE_ADDRESS_0,
                                &pci_ior_addr);
      pcibios_read_config_dword(pci_bus, pci_device_fn, PCI_BASE_ADDRESS_1,
                                &pci_pbor_addr);
      pci_ior_addr  &= PCI_BASE_ADDRESS_IO_MASK;
      pci_pbor_addr &= PCI_BASE_ADDRESS_IO_MASK;
      
      cc_pci_hdr->irq   = pci_irq_line;
      cc_pci_hdr->base1 = pci_ior_addr;
      cc_pci_hdr->base2 = pci_pbor_addr;

      cc_reg->csr  += pci_pbor_addr;
      cc_reg->cnaf += pci_pbor_addr;
      cc_reg->tcr  += pci_pbor_addr;
      cc_reg->srr  += pci_pbor_addr;

      cc_reg->fifo += pci_ior_addr;
      cc_reg->mwa  += pci_ior_addr;
      cc_reg->mwtc += pci_ior_addr;
      cc_reg->mra  += pci_ior_addr;
      cc_reg->mrtc += pci_ior_addr;
      cc_reg->ics  += pci_ior_addr;
      cc_reg->bmcs += pci_ior_addr;
      
      printk("K2915 card at ior = %#x, pbor = %#x, irq = %d\n",
		pci_ior_addr, pci_pbor_addr, pci_irq_line);
      
      break;
    }		
  }
  else
    value = -1;
  
  return value;
}
