/*
 * File Name     : camlib.c 
 * Version       : v1.0
 * Introduction  : for ISA/K2927-Z1A and PCI/K2915
 * Creation Date : Dec. 1996
 * Author        : M.Haseno
 *               : Nagasaki Institute of Applied Science
 *
 * modified      : 3-Mar-1998 Carl Timmer, Jefferson Lab
 *		   1) struct naf_data now has xq as unsigned int
 *		      instead of short, so no information is lost.
 *		   2) Functions CCMDSTAT(int*),
 *				CPPOLL (int *),
 *		      		CREGDUMP (void),
 *				CENINT (void),
 *				CDISINT (void) were added.
 *
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <sys/errno.h>

#include "camlib.h"
/* for debugging */
/*#define DEBUGCLIB 1*/

/* Local Storage */
static int cc_busy = CC_K_READY;
static int cc_path;
static int cc_cur_crate;
static int cc_cur_branch;
static int cc_event_count;

struct naf_data {
  short mode;
  short naf;
  unsigned int data;
  unsigned int xq;
};

/* OPEN */
int
cam_open_ (void)
{
  return COPEN ();
}
int
camopn_ (void)
{
  return COPEN ();
}
int
copen_ (void)
{
  return COPEN ();
}
int
COPEN (void)
{
  if (cc_busy == CC_K_BUSY)
    return -EBUSY;
  else
    cc_busy = CC_K_BUSY;
    
  cc_path = open ("/dev/cc0", O_RDWR);
  if (cc_path == -1)
    return -ENODEV;
  
  cc_cur_branch  = 0;
  cc_cur_crate   = 0;
  cc_event_count = 0;
  
#ifdef DEBUGCLIB
printf("COPEN: opened things\n");
#endif 
 
  return 0;
}

/* CLOSE */
int
cam_close_ (void)
{
  return CCLOSE ();
}
int
camcls_ (void)
{
  return CCLOSE ();
}
int
cclose_ (void)
{
  return CCLOSE ();
}
int
CCLOSE (void)
{
  cc_busy = CC_K_READY;
  if ((close (cc_path)) == -1)
    return errno;
    
  return 0;
}

/* NAF */
int
naf_ (int *n, int *a, int *f)
{
  return NAF (*n, *a, *f);
}
int
NAF (int n, int a, int f)
{
  return (0x3fff & (n << 9 | a << 5 | f));
}

/* CRESET */
int
creset_ (void)
{
  return CRESET ();
}
int
CRESET (void)
{
  int status;
  unsigned long arg;

  status = ioctl (cc_path, CCIOC_RESET, arg);
  if (status != 0)
    printf ("CRESET : status=%d\n", status);
    
  return status;
}

/* CCMDSTAT */
int
ccmdstat_ (int *cmdstat)
{
  return CCMDSTAT (cmdstat);
}
int
CCMDSTAT (int *cmdstat)
{
  /* Put status of last cmd (x,q,error,
   * timeout bits from csr) into cmdstat
   */
  
  int status;
  
  status = ioctl (cc_path, CCIOC_CMD_STAT, cmdstat);
  if (status != 0)
    printf ("CCMDSTAT: ioctl error, status = %d\n", status);

  return status;
}


/* CPPOLL */
int
cppoll_ (int *poll)
{
  return CCMDSTAT (poll);
}
int
CPPOLL (int *poll)
{
  /* Do parallel poll & put status into "poll" */
  
  int status;
  
  status = ioctl (cc_path, CCIOC_PPOLL, poll);
  if (status != 0)
    printf ("CPPOLL: ioctl error, status = %d\n", status);

  return status;
}

/* CREGDUMP */
int
cregdump_ (void)
{
  return CREGDUMP();
}
int
CREGDUMP (void)
{
  /*
   * Do a dump of the k2915's registers with printk. 
   * Do a "dmesg" command to see results.
   */
  
  int status, dummy;
  
  status = ioctl (cc_path, CCIOC_REG_DUMP, &dummy);
  if (status != 0)
    printf ("CREGDUMP: ioctl error, status = %d\n", status);

  return status;
}

/* CSETBR */
int
csetbr_ (int *branch)
{
  return CSETBR (*branch);
}
int
CSETBR (int branch)
{
  int status;

  if (branch >= 0 && branch < CC_K_MAX_BRANCH) {
    cc_cur_branch = branch;
    status = ioctl (cc_path, CCIOC_SET_BRANCH, (unsigned long) branch);
    if (status != 0)
      printf ("CSETBR : status=%d\n", status);
  }
  else
    return -EINVAL;
    
  return status;
}

/* CSETCR */
int
csetcr_ (int *crate)
{
  return CSETCR (*crate);
}
int
CSETCR (int crate)
{
  int status;

  if (crate >= 0 && crate < CC_K_MAX_CRATE) {
    cc_cur_crate = crate;
    status = ioctl (cc_path, CCIOC_SET_CRATE, (unsigned long) crate);
    if (status != 0)
      printf ("CSETCR : status=%d\n", status);
  }
  else
    return -EINVAL;
    
  return status;
}

/* CGENZ */
int
cgenz_ (void)
{
  return CGENZ ();
}
int
CGENZ (void)
{
  int status, q, x;
  short i;
  
  /* read Status register of crate */
  status = CAMACW (NAFGEN (30, 0, 1), &i, &q, &x);
  if (status != 0)
    return status;
    
  /* OR Status with generate_Z (initialize) bit */
  i |= GEN_Z;
  
  /* write Status to crate */
  status = CAMACW (NAFGEN (30, 0, 17), &i, &q, &x);
  if (status != 0)
    return status;
  return 0;
}

/* CGENC */
int
cgenc_ (void)
{
  return CGENC ();
}
int
CGENC (void)
{
  int status, q, x;
  short i;

  status = CAMACW (NAFGEN (30, 0, 1), &i, &q, &x);
  if (status != 0)
    return status;
     
  i |= GEN_C;
  
  status = CAMACW (NAFGEN (30, 0, 17), &i, &q, &x);
  if (status != 0)
    return status;

  return 0;
}

/* CSETI */
int
cseti_ (void)
{
  return CSETI ();
}
int
CSETI (void)
{
  int status, q, x;
  short i;

  status = CAMACW (NAFGEN (30, 0, 1), &i, &q, &x);
  if (status != 0)
    return status;
    
  i |= INHIBIT;
  
  status = CAMACW (NAFGEN (30, 0, 17), &i, &q, &x);
  if (status != 0)
    return status;

  return 0;
}

/* CREMI */
int
cremi_ (void)
{
  return CSETI ();
}
int
CREMI (void)
{
  int status, q, x;
  short i;

  status = CAMACW (NAFGEN (30, 0, 1), &i, &q, &x);
  if (status != 0)
    return status;
    
  i &= ~INHIBIT;
  
  status = CAMACW (NAFGEN (30, 0, 17), &i, &q, &x);
  if (status != 0)
    return status;

  return 0;
}

/* CAMACW (16bit CAMAC data transfer) */
int
camacw_ (int *naf, unsigned short *data, int *q, int *x)
{
  return CAMACW (*naf, data, q, x);
}
int
CAMACW (int naf, unsigned short *data, int *q, int *x)
{
  unsigned int status;
  struct naf_data command;

  command.mode = SINGLE16;
  command.naf = naf;
  
  if ((CC_NAFRK_F16 & naf) && !(CC_NAFRK_F08 & naf))  /*  Write */
    command.data = (int)*data;
  
  status = ioctl(cc_path, CCIOC_SINGLE, &command);
  if (status != 0){
    printf ("CAMACW : status=%d\n", status);
    return (int) status;
  }

  if (!(CC_NAFRK_F16 & naf) && !(CC_NAFRK_F08 & naf)) /*  Read */
    *data = (unsigned short)command.data;

  status = command.xq;
  if (status & NOQ)
    *q = 0;
  else
    *q = 1;
  if (status & NOX)
    *x = 0;
  else
    *x = 1;
    
#ifdef DEBUGCLIB
printf("CAMACW: naf = %#x, data = %#x, cmd.xq = %#x, q = %d, x = %d\n", naf, *data, status, *q, *x);  
#endif

  return 0;
}

/* CAMAC (24bit CAMAC data transfer) */
int
camac_ (int *naf, int *data, int *q, int *x)
{
  return CAMAC (*naf, data, q, x);
}
int
CAMAC (int naf, int *data, int *q, int *x)
{
  unsigned int status;
  struct naf_data command;

  command.mode = SINGLE24;
  command.naf = naf;

  if ((CC_NAFRK_F16 & naf) && !(CC_NAFRK_F08 & naf))	/* Write */
    command.data = *data;
  
  status = ioctl(cc_path, CCIOC_SINGLE, &command);
  if (status != 0){
    printf ("CAMAC : status=%d\n", status);
    return (int) status;
  }

  if (!(CC_NAFRK_F16 & naf) && !(CC_NAFRK_F08 & naf))	/* Read */
    *data = command.data;

  status = command.xq;
  if (status & NOQ)
    *q = 0;
  else
    *q = 1;
  if (status & NOX)
    *x = 0;
  else
    *x = 1;
    
#ifdef DEBUGCLIB
printf("CAMAC: naf = %#x, data = %#x, cmd.xq = %#x, q = %d, x = %d\n", naf, *data, status, *q, *x);  
#endif

  return 0;
}

/* CENINT 
 *  Enable interrupts to isr, "cc_interrupt".
 *  This writes RFS_IENA|(inl(reg->csr) & RFS_MASK)
 *  to csr register
 */
int
cenint_ (void)
{
  CENINT ();
}
int
CENINT (void)
{
  int status;
  unsigned long arg;

  status = ioctl (cc_path, CCIOC_ENABLE_INT, arg);
  if (status != 0){
    printf ("CENINT : status=%d\n", status);
    return status;
  }

  return 0;
}

/* CDISINT 
 *  Disable interrupts.
 *  This writes ~RFS_IENA & inl(reg->csr) & RFS_MASK
 *  to csr register.
 */
int
cdisint_ (void)
{
  CDISINT ();
}
int
CDISINT (void)
{
  int status;
  unsigned long arg;

  status = ioctl (cc_path, CCIOC_DISABLE_INT, arg);
  if (status != 0){
    printf ("CENINT : status=%d\n", status);
    return status;
  }

  return 0;
}

/* Cam_EnableLAM,CENLAM,CELAM */
int
cam_enablelam_ (int *mask)
{
  return CENLAM (*mask);
}
int
cenlam_ (int *mask)
{
  return CENLAM (*mask);
}
int
celam_ (int *mask)
{
  return CENLAM (*mask);
}
int
CENLAM (int mask)
{
  int status;

  status = ioctl (cc_path, CCIOC_ENABLE_LAM, (unsigned long) mask);
  if (status != 0){
    printf ("CENLAM : status=%d\n", status);
    return status;
  }

  return 0;
}

/* Cam_DisableLAM,CDSLAM,CDLAM */
int
cam_disablelam_ (void)
{
  return CDSLAM ();
}
int
cdslam_ (void)
{
  return CDSLAM ();
}
int
cdlam_ (void)
{
  return CDSLAM ();
}
int
CDSLAM (void)
{
  int status;
  long arg;

  status = ioctl (cc_path, CCIOC_DISABLE_LAM, arg);
  if (status != 0){
    printf ("CDSLAM : status=%d\n", status);
    return status;
  }

  return 0;
}

/* Cam_WaitLAM,CWTLAM,CWLAM */
int
cam_waitlam_ (int *timeout)
{
  return CWTLAM (*timeout);
}
int
cwtlam_ (int *timeout)
{
  return CWTLAM (*timeout);
}
int
cwlam_ (int *timeout)
{
  return CWTLAM (*timeout);
}
int
CWTLAM (int timeout)
{
  int status;

  status = ioctl (cc_path, CCIOC_WAIT_LAM, (unsigned long) timeout);
  if (status != 0)
    printf ("CWTLAM : status=%d\n", status);
  return status;
}


/* CDMAW */
int
CDMAW (int mode, int naf, unsigned short *buf, int size, int *retlen, int *err)
{
  int status;
  struct naf_data command;
/*
  if(size > 0x20000){
    printf("Buffer is too long...\n");
    return -1;
  }
*/
  command.mode = (short)(CC_K_MODE_BLOCK | CC_K_MODE_WS16 | mode);  
  command.naf  = (short)naf;

  status = ioctl (cc_path, CCIOC_SETUP_DMA, &command);
  if (status != 0){
    printf ("CDMAW : status=%d\n", status);
    return status;
  }

  if ((CC_NAFRK_F16 & naf) && !(CC_NAFRK_F08 & naf)){ /* Write */
    *err = write (cc_path,(char *)buf,size);
    if(*err !=0)
      printf("CDMAW:Error status=%d\n", *err);
    *retlen = size;
  }
  else if (!(CC_NAFRK_F16 & naf) && !(CC_NAFRK_F08 & naf)){   /* Read */
    *err = read (cc_path,buf,size);
    if(*err !=0)
      printf("CDMAW:Error status=%d\n", *err);
    *retlen = size;
  }
  else
    return -1;
  return 0;
}

/* CDMAL */
int
CDMAL (int mode, int naf, int *buf, int size, int *retlen, int *err)
{
  int status;
  struct naf_data command;
/*
  if(size > 0x20000){
    printf("Buffer is too long...\n");
    return -1;
  }
*/
  command.mode = (short)(CC_K_MODE_BLOCK | CC_K_MODE_WS24 | mode);  
  command.naf  = (short)naf;

  status = ioctl (cc_path, CCIOC_SETUP_DMA, &command);
  if (status != 0){
    printf ("CDMAL : status=%d\n", status);
    return status;
  }

  if ((CC_NAFRK_F16 & naf) && !(CC_NAFRK_F08 & naf)){ /* Write */
    *err = write (cc_path,(char *)buf,size);
    if(*err !=0)
      printf("CDMAW:Error status=%d\n", *err);
    *retlen = size;
  }
  else if (!(CC_NAFRK_F16 & naf) && !(CC_NAFRK_F08 & naf)){   /* Read */
    *err = read (cc_path,(char *)buf,size);
    if(*err !=0)
      printf("CDMAW:Error status=%d\n", *err);
    *retlen = size;
  }
  else
    return -1;
  return 0;
}



