/*-----------------------------------------------------------------------------
 * Copyright (c) 1998 Southeastern Universities Research Association,
 *                    Thomas Jefferson National Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 * TJNAF Data Acquisition Group, 12000 Jefferson Ave., Newport News, VA 23606
 * Email: coda@cebaf.gov  Tel: (757)269-5130  Fax: (757)269-5800
 *-----------------------------------------------------------------------------
 * 
 * Description:
 *   An implementation of the CAMAC standard routines in C,
 *   using C argument passing conventions. Functions written
 *   as calls to Kinetic Systems 2915 library.
 *	
 *	
 * Author:  Carl Timmer, TJNAF Data Acquisition Group
 *
 * Revision History:
 *   $Log: libks2915.c,v $
 *	  Revision 1.1  1998/03/04  10:20:00  timmer
 *	  Initial revision
 *
 */

/* Functions (all return void):
 *
 * cdreg(int *ext,int b,int c,int n,int a)   define external address 
 * cssa(int f,int ext,short *dat,int *q)     short (16) single action 
 * cfsa(int f,int ext,int *dat,int *q)       long  (24) single action 
 * cccz(int ext)			 crate initialize 
 * cccc(int ext)			 crate clear 
 * ccci(int ext,int l)			 set/reset crate inhibit 
 * ctci(int ext,int *l)			 test crate inhibit 
 * cccd(int ext,int l)			 ena/disable crate demand 
 * ctcd(int ext,int *l)			 test crate demand enabled
 * ccgl(int lam,int l)			 ena/disable graded lam (NON STANDARD)
 * ctgl(int ext,int *l)			 test graded lam
 * cdlam(int *lam,int b,int c,int n,int a,int inta[]) 	 define lam 
 * cclc(int lam)			 clear lam (dataless function)
 * cclm(int lam,int l)			 alternative lam clears 
 * ctlm(int lam,int *l)			 test lam 
 * ccinit(int b)			 initialize branch 
 * cclnk(int lam,void *())		 link routine to lam 
 * cculk(int lam,void *())		 unlink routine from lam 
 * ccrgl(int lam)			 re-enable graded lam (@end of isr)
 * cclwt(int lam)			 lam wait 
 * ctstat(int *istat)			 test status of previous operation 
 */

#include <stdio.h>
#include "ks2915.h"
#include "camlib.h"
typedef void (*VOIDFUNCPTR) ();

void ccinit(int b)
{
  int i, q, x, cmdstat;
  unsigned short data;
  
  if (b<0 || b>7) {
    printf("ccinit: 0 <= branch <= 7");
    exit(1);
  }
  
  if (COPEN()) {
    perror("Open error: ");
    exit(1);
  }
  
  /* only branch 0 is supported */
  CSETBR(0);
  
  for (i=0; i<8; i++) {
    CSETCR(i);
    /* Check to see if crate is there (no timeout)
     * by reading register that's always readable
     * (even when off-line!).
     */
    CAMACW(NAF(30,0,1), &data, &q, &x);
    CCMDSTAT(&cmdstat);
    if(cmdstat & (NAF_TIMEOUT | PBUS_TIMEOUT)) 
      continue;
      
    if(data & OFFLINE) {
      printf("Crate #%d is offline\n",i);
      continue;
    }
      
    CGENC();
    CGENZ();
    CREMI();
  }
  
  CSETCR(0);
  printf("ccinit done\n");
}

void cdreg(int *ext,int b,int c,int n,int a)
{
  if( (b<0||b>7) ||(c<0||c>7)||
      (n<0||n>31)||(a<0||a>15) )
    *ext = -1;
  else
    CDREG(ext,b,c,n,a);
printf("cdreg: ext = %d\n", *ext);
}

void cfsa(int f,int ext,unsigned long *dat,unsigned long *q)
{
  int naf,x,crate;
  
  GETC(ext, &crate);
  CSETCR(crate);
  GETNAF(ext,f,&naf);
  /* printf("cfsa: pass naf=%d, data=%#x\n", naf, *dat); */
  CAMAC(naf, (int *) dat, (int *) q, &x);
  /* printf("cfsa: retn data=%#x, q = %d, x = %d\n", *dat, *q, x); */
}

void cccz(int ext)
{
  int crate;
  
  GETC(ext, &crate);
  CSETCR(crate);
  CGENZ();
}

void cccc(int ext)
{
  int crate;
  
  GETC(ext, &crate);
  CSETCR(crate);
  CGENC();
}

void ccci(int ext,int l)
{
  int crate;
  
  GETC(ext, &crate);
  CSETCR(crate);
  (l==1)? CSETI() : CREMI();
}

void ctci(int ext,int *l)
{
  int q, x, crate;
  short i;

  GETC(ext, &crate);
  CSETCR(crate);
  CAMACW( NAFGEN(30,0,1), &i, &q, &x );
  *(l) = (i>>2) & 1;
}

void cccd(int ext,int l)
{
  int q, x, crate;
  short i;

  GETC(ext, &crate);
  CSETCR(crate);
  CAMACW( NAFGEN(30,0,1), &i, &q, &x );
  i = (l ? (i|0x100) : (i & ~0x100));
  CAMACW( NAFGEN(30,0,17), &i, &q, &x );
}

void ctcd(int ext,int *l)
{
  int q, x, crate;
  short i;

  GETC(ext, &crate);
  CSETCR(crate);
  CAMACW( NAFGEN(30,0,1), &i, &q, &x );
  *(l) = ((i & 0x100) ? 1 : 0);
}

void ccgl(int lam,int l)
{
  int i, q, x, crate;

  GETC(lam, &crate);
  CSETCR(crate);
  
  /* read crate controller lam mask */
  CAMAC( NAFGEN(30,13,1), &i, &q, &x );
  if (l) {
    i = i |  (1<<(((lam>>9)-1) & 31));
  } else {
    i = i & ~(1<<(((lam>>9)-1) & 31));
  }
  /* write lam mask */
  CAMAC( NAFGEN(30,13,17), &i, &q, &x );
}
  
void ctgl(int ext,int *l)
{
  int q, x, crate;
  short i;

  GETC(ext, &crate);
  CSETCR(crate);
  CAMACW( NAFGEN(30,0,1), &i, &q, &x );
  *(l) = (i>>15) & 1;
}

void cdlam(int *ext,int b,int c,int n,int a,int inta[2])
{
  /* ignore inta[] */
  cdreg(ext,b,c,n,a);
}
     
void cclm(int lam,int l)
{
  unsigned long i;
  int crate;
  
  GETC(lam, &crate);
  CSETCR(crate);
  
  if(l) {
    cfsa(26,lam,&i,&i);
  } else {
    cfsa(24,lam,&i,&i);
  }
}

void cclc(int lam)
{ 
  unsigned long i;
  cfsa(10,lam,&i,&i);
}

void ctlm(int lam,int *l)
{
  unsigned long i, j;
  cfsa(8,lam,&i,&j);
  *l = j;
}

/* enable interrupt & wait for one (1min timeout) */
void ccrgl(int lam)
{
  int mask, crate;

  GETC(lam, &crate);
  CSETCR(crate);
  CWTLAM(0);
}

void cclnk(int lam,VOIDFUNCPTR label)
{
  /* This routine unneccesary in Linux */
}

void cculk(int lam,VOIDFUNCPTR label)
{
  /* This routine unneccesary in Linux */
}

void cclwt(int lam)
{
  /* This routine unneccesary in Linux */
}

void ctstat(int *istat)
{
  /* Check status of last cmd: x, q, error,& timeout bits */
  
  int status, dat, x, q, d, e;
  
  status = CCMDSTAT(&dat);
  if (status != 0) {
    printf ("ctstat: error in CCMDSTAT\n");
    return;
  }

  /* x=q=1 -> d=0; x=1,q=0 -> 1; x=0,q=1 -> 2; x=q=0 -> 3 */
  x = ((dat & NOX)>>17) ? 2 : 0;
  q = ((dat & NOQ)>>16) ? 1 : 0;
  d = x + q;

  /* error bit */
  e = (dat&ERR)>>31;

  /* parallel bus timeout */
  if (dat & PBUS_TIMEOUT)
    e += 1;

  /* naf timeout */
  if (dat & NAF_TIMEOUT)
    e += 2;

  *(istat) = 4*e + d;
} 

void clpoll(lam)
{
  int l;
  for (l=0;!l;) ctlm(lam,&l);
}

/*********************** Multiple Action Routines ***************************/

void cfga(int f[],int exta[],int intc[],int qa[],int cb[])
{
  int i;
  for (i=0;i<cb[0];i++) {
    cfsa(f[i],exta[i],(unsigned long *) &(intc[i]),(unsigned long *) &(qa[i]));
  }
  cb[1] = cb[0];
}

void cfmad(int f,int extb[],int intc[],int cb[])
{
  int naf, err, retlen, crate;
  
  GETC(extb[0], &crate);
  CSETCR(crate);
  GETNAF(extb[0],f,&naf);  
  CDMAL(CC_Q_SCAN, naf, intc, cb[0], &cb[1], &err);
}

void cfubc(int f,int ext,int intc[],int cb[])
{
  int naf, err, retlen, crate;
  
  GETC(ext, &crate);
  CSETCR(crate);
  GETNAF(ext,f,&naf);  
  CDMAL(CC_Q_STOP, naf, intc, cb[0], &cb[1], &err);
}

void cfubr(int f,int ext,int intc[],int cb[])
{
  int naf, err, retlen, crate;
  
  GETC(ext, &crate);
  CSETCR(crate);
  GETNAF(ext,f,&naf);  
  CDMAL(CC_Q_REPEAT, naf, intc, cb[0], &cb[1], &err);
}

void cfubl(int f,int ext,int intc[],int cb[])
{
  int i;
  unsigned long q;
  
  for (i=0;i<cb[0];i++) {
    clpoll(cb[2]);
    cfsa(f,ext,(unsigned long *) &(intc[i]),&q);
    if (!q) {
      i--;
      break;
    }
  }
  cb[1]=i+1;
}

/************************ Short Data-Word Transfers ************************/

void cssa(int f,int ext,unsigned short *dat,unsigned long *q)
{
  int naf, x, crate;
  
  GETC(ext, &crate);
  CSETCR(crate);
  GETNAF(ext,f,&naf);  
  /* printf("cssa: pass naf=%d, data=%#x\n", naf, *dat); */
  CAMACW(naf, dat, (int *)q, &x);
  /* printf("cssa: retn data=%#x, q = %d, x = %d\n", *dat, *q, x); */
}

void csga(int f[],int exta[],unsigned short intc[],int qa[],int cb[])
{
  int i;
  for (i=0;i<cb[0];i++) {
    cssa(f[i],exta[i],&(intc[i]),(unsigned long *) &(qa[i]));
  }
  cb[1] = cb[0];
}

void csmad(int f,int extb[],unsigned short intc[],int cb[])
{
  int naf, err, retlen, crate;
  
  GETC(extb[0], &crate);
  CSETCR(crate);
  GETNAF(extb[0],f,&naf);  
  CDMAW(CC_Q_SCAN, naf, intc, cb[0], &cb[1], &err);
}

void csubc(int f,int ext,unsigned short intc[],int cb[])
{
  int naf, err, retlen, crate;
  
  GETC(ext, &crate);
  CSETCR(crate);
  GETNAF(ext,f,&naf);  
  CDMAW(CC_Q_STOP, naf, intc, cb[0], &cb[1], &err);
}

void csubr(int f,int ext,unsigned short intc[],int cb[])
{
  int naf, err, retlen, crate;
  
  GETC(ext, &crate);
  CSETCR(crate);
  GETNAF(ext,f,&naf);  
  CDMAW(CC_Q_REPEAT, naf, intc, cb[0], &cb[1], &err);
}

void csubl(int f,int ext,unsigned short intc[],int cb[])
{
  int i;
  unsigned long q;
  
  for (i=0;i<cb[0];i++) {
    clpoll(cb[2]);
    cssa(f,ext,&(intc[i]),&q);
    if (!q) {
      i--;
      break;
    }
  }
  cb[1]=i+1;
}


